import os
import shutil
from typing import Callable, Optional, Tuple, Iterable, Union
from warnings import warn
from dataclasses import dataclass
import h5py
from pathlib import Path
import hashlib
import numpy as np
import logging


def save_to_cache(data, targets, file_path: Union[str, Path]) -> None:
    """
    Save data to caching path on disk in an hdf5 file. Can deal with data
    that is a dictionary.
    Args:
        data: numpy ndarray-like or a list thereof.
        targets: same as data, can be None.
        file_path: caching file path.
    """
    with h5py.File(file_path, "w") as f:
        for name, data in zip(["data", "target"], [data, targets]):
            if type(data) != tuple: data = (data,)
            # can be events, frames, imu, gps, target etc.
            for i, data_piece in enumerate(data):
                if type(data_piece) == dict:
                    for key, item in data_piece.items():
                        f.create_dataset(f"{name}/{i}/{key}", data=item, compression='lzf' if type(item) == np.ndarray else None)
                else:
                    f.create_dataset(f"{name}/{i}", data=data_piece, compression='lzf' if type(data_piece) == np.ndarray else None)


def load_from_cache(file_path: Union[str, Path]) -> Tuple:
    """
    Load data from file cache, separately for (data) and (targets). Can assemble dictionaries back together.
    Args:
        file_path: caching file path.
    Returns:
        data, targets
    """
    data_list = []
    target_list = []
    with h5py.File(file_path, "r") as f:
        for name, _list in zip(["data", "target"], [data_list, target_list]):
            for index in f[name].keys():
                if hasattr(f[f"{name}/{index}"], "keys"):
                    data = {key: f[f"{name}/{index}/{key}"][()] for key in f[f"{name}/{index}"].keys()}
                else:
                    data = f[f"{name}/{index}"][()]
                _list.append(data)
    if len(data_list) == 1: data_list = data_list[0]
    if len(target_list) == 1: target_list = target_list[0]
    return data_list, target_list


@dataclass
class CachedDataset:
    """
    CachedDataset caches the data samples to the hard drive for subsequent reads, thereby potentially improving data loading speeds.
    If dataset is None, then the length of this dataset will be inferred from the number of files in the caching folder. Pay
    attention to the cache path you're providing, as CachedDataset will simply check if there is a file present with the index that
    it is looking for. When using train/test splits, it is wise to also take that into account in the cache path.

    Parameters:
        dataset:
            Dataset to be cached. Can be None, if only files in cache_path should be used.
        cache_path:
            The preferred path where the cache will be written to and read from.
        reset_cache:
            When True, will clear out the cache path during initialisation. Default is False
        transform:
            Transforms to be applied on the data
        target_transform:
            Transforms to be applied on the label/targets
        num_copies:
            Number of copies of each sample to be cached.
            This is a useful parameter if the dataset is being augmented with slow, random transforms.
    """

    dataset: Iterable
    cache_path: str
    reset_cache: bool = False
    transform: Optional[Callable] = None
    target_transform: Optional[Callable] = None
    num_copies: int = 1

    def __post_init__(self):
        super().__init__()
        # Create cache directory
        if not os.path.isdir(self.cache_path):
            os.makedirs(self.cache_path)
        if self.reset_cache:
            shutil.rmtree(self.cache_path)
            os.makedirs(self.cache_path)
        if self.dataset is None:
            self.n_samples = len([name for name in os.listdir(self.cache_path) if os.path.isfile(name) and name.endswith('.hdf5')]) // self.num_copies
        else:
            self.n_samples = len(self.dataset)

    def __getitem__(self, item) -> (object, object):
        copy = np.random.randint(self.num_copies)
        file_path = os.path.join(self.cache_path, f"{item}_{copy}.hdf5")
        try:
            data, targets = load_from_cache(file_path)
        except (FileNotFoundError, OSError) as _:
            logging.info(
                f"Data {item}: {file_path} not in cache, generating it now", stacklevel=2
            )
            data, targets = self.dataset[item]
            save_to_cache(data, targets, file_path=file_path)
            # format might change during save to hdf5,
            # i.e. tensors -> np arrays
            data, targets = load_from_cache(file_path)

        if self.transform:
            data = self.transform(data)
        if self.target_transform:
            targets = self.target_transform(targets)
        return data, targets

    def __len__(self):
        return self.n_samples



import os
import shutil
import h5py
import numpy as np
from pathlib import Path
from tonic import datasets
import tonic.transforms as transforms
from tonic import CachedDataset


def test_caching_pokerdvs():
    dataset = datasets.POKERDVS(save_to="./data", train=False)
    dataset = CachedDataset(dataset, cache_path="./cache/cache1")
    for data, label in dataset:
        print(data.shape, label)

def test_caching_transforms():
    sensor_size = datasets.POKERDVS.sensor_size
    preprocess = transforms.Compose(
        [transforms.Downsample(time_factor=1, spatial_factor=1)]
    )
    augmentation = transforms.Compose(
        [transforms.Downsample(time_factor=1, spatial_factor=1)]
    )
    dataset = datasets.POKERDVS(save_to="./data", train=True, transform=preprocess)

    dataset_cached = CachedDataset(dataset, cache_path="./cache/cache2", transform=augmentation, num_copies=4)

    for (data, label), (data2, label2) in zip(dataset, dataset_cached):
        assert (data == data2).all()
        assert label == label2

def test_caching_reset():
    cache_path = Path('./cache/test3')
    if cache_path.exists():
        shutil.rmtree(cache_path)

    # simulate outdated cache file
    old_file = cache_path / '0_0.hdf5'
    dummy_content = np.zeros((3,3))
    os.makedirs(cache_path)
    with h5py.File(old_file, "w") as f:
        f.create_dataset(f"data/0", data=dummy_content)
        f.create_dataset(f"target/0", data=dummy_content)
    
    dataset = datasets.POKERDVS(save_to="./data", train=False)

    # load first sample from cache, which will load the old file
    dataset_cached = CachedDataset(dataset, cache_path=cache_path, reset_cache=False)
    data, target = dataset_cached[0]
    assert (data == dummy_content).all()
    
    # when using reset_cache=True, the folder will be reset and the right file cached
    dataset_cached = CachedDataset(dataset, cache_path=cache_path, reset_cache=True)
    data, target = dataset_cached[0]
    assert data.shape != dummy_content.shape
    assert data.shape[0] > 100



from . import datasets, transforms, collation, io, slicers, utils
from .version import version as __version__
from .dataset import Dataset
from .cached_dataset import CachedDataset
from .sliced_dataset import SlicedDataset



import os
import shutil
import glob
import numpy as np
import loris
import numpy
from tonic.dataset import Dataset
from tonic.download_utils import extract_archive


class NavGesture(Dataset):
    """NavGesture dataset <https://www.neuromorphic-vision.com/public/downloads/navgesture/>. Events have (txyp) ordering.
    ::

        @article{maro2020event,
          title={Event-based gesture recognition with dynamic background suppression using smartphone computational capabilities},
          author={Maro, Jean-Matthieu and Ieng, Sio-Hoi and Benosman, Ryad},
          journal={Frontiers in neuroscience},
          volume={14},
          pages={275},
          year={2020},
          publisher={Frontiers}
        }

    Parameters:
        save_to (string): Location to save files to on disk.
        walk_subset (bool): Choose either NavGesture-sit (default) or NavGesture-walk dataset. No train/test split provided.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
    """

    base_url = "https://www.neuromorphic-vision.com/public/downloads/navgesture/"
    sit_filename = "navgesture-sit.zip"
    walk_filename = "navgesture-walk.zip"
    sit_url = base_url + sit_filename
    walk_url = base_url + walk_filename
    sit_md5 = "1571753ace4d9e0946e6503313712c22"
    walk_md5 = "5d305266f13005401959e819abe206f0"

    classes = ["swipe down", "swipe up", "swipe left", "swipe right", "select", "home"]
    class_codes = ["do", "up", "le", "ri", "se", "ho"]
    int_classes = dict(zip(class_codes, range(len(class_codes))))
    sensor_size = (304, 240, 2)
    dtype = np.dtype([("t", "<u8"), ("x", "<u2"), ("y", "<u2"), ("p", "?")])
    ordering = dtype.names

    def __init__(
        self, save_to, walk_subset=False, transform=None, target_transform=None
    ):
        super(NavGesture, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )

        self.walk_subset = walk_subset

        if walk_subset:
            self.url = self.walk_url
            self.file_md5 = self.walk_md5
            self.filename = self.walk_filename
            self.folder_name = "navgesture-walk"
        else:
            self.url = self.sit_url
            self.file_md5 = self.sit_md5
            self.filename = self.sit_filename
            self.folder_name = "navgesture-sit"

        if not self._check_exists():
            self.download()
            data_folder = os.path.join(self.location_on_system, self.folder_name)
            # normally zips contain a top-level folder where we can extract to,
            # but here we have to create and move the data into it manually
            os.makedirs(data_folder, exist_ok=True)
            pattern = "/user*.zip"
            files = glob.glob(self.location_on_system + pattern)
            for file in files:
                file_name = os.path.basename(file)
                shutil.move(file, os.path.join(data_folder, file_name))
            for path, dirs, files in os.walk(data_folder):
                dirs.sort()
                for file in files:
                    if file.startswith("user") and file.endswith("zip"):
                        extract_archive(os.path.join(data_folder, file))

        for path, dirs, files in os.walk(self.location_on_system):
            dirs.sort()
            files.sort()
            for file in files:
                if file.endswith("dat"):
                    self.data.append(path + "/" + file)
                    self.targets.append(self.int_classes[file[7:9]])

    def __getitem__(self, index):
        """
        Returns:
            a tuple of (events, target) where target is the index of the target class.
        """
        events, target = (
            loris.read_file(self.data[index])["events"],
            self.targets[index],
        )
        events = np.lib.recfunctions.rename_fields(events, {'ts': 't', 'is_increase': 'p'})

        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        return len(self.data)

    def _check_exists(self):
        return self._is_file_present() and self._folder_contains_at_least_n_files_of_type(
            304, ".dat"
        )



import os
import loris
import numpy as np
from tonic.dataset import Dataset
from tonic.download_utils import extract_archive


class NCARS(Dataset):
    """N-Cars dataset <https://www.prophesee.ai/dataset-n-cars-download/>. Events have (txyp) ordering.
    ::

        @inproceedings{sironi2018hats,
          title={HATS: Histograms of averaged time surfaces for robust event-based object classification},
          author={Sironi, Amos and Brambilla, Manuele and Bourdis, Nicolas and Lagorce, Xavier and Benosman, Ryad},
          booktitle={Proceedings of the IEEE Conference on Computer Vision and Pattern Recognition},
          pages={1731--1740},
          year={2018}
        }

    Parameters:
        save_to (string): Location to save files to on disk.
        train (bool): If True, uses training subset, otherwise testing subset.
        transform (callable, optional): A callable of transforms to apply to the data.
        target_transform (callable, optional): A callable of transforms to apply to the targets/labels.
    """

    url = "http://www.prophesee.ai/resources/Prophesee_Dataset_n_cars.zip"
    filename = "Prophesee_Dataset_n_cars.zip"
    train_filename = "n-cars_train.zip"
    test_filename = "n-cars_test.zip"
    file_md5 = "553ce464d6e5e617b3c21ce27c19368e"
    classes = ["background", "car"]

    class_dict = {"background": 0, "cars": 1}

    sensor_size = None  # different for every recording
    minimum_y_value = 140
    dtype = np.dtype([("t", "<u8"), ("x", "<u2"), ("y", "<u2"), ("p", "?")])
    ordering = "txyp"

    def __init__(self, save_to, train=True, transform=None, target_transform=None):
        super(NCARS, self).__init__(
            save_to, transform=transform, target_transform=target_transform
        )

        if train:
            self.folder_name = "train"
        else:
            self.folder_name = "test"

        if not self._check_exists():
            self.download()
            extract_archive(os.path.join(self.location_on_system, self.train_filename))
            extract_archive(os.path.join(self.location_on_system, self.test_filename))

        file_path = os.path.join(self.location_on_system, self.folder_name)
        for path, dirs, files in os.walk(file_path):
            dirs.sort()
            for file in files:
                if file.endswith("dat"):
                    self.data.append(path + "/" + file)
                    self.targets.append(self.class_dict[os.path.basename(path)])

    def __getitem__(self, index):
        """
        Returns:
            a tuple of (events, target) where target is the index of the target class.
        """
        events = loris.read_file(self.data[index])["events"]
        events = np.lib.recfunctions.rename_fields(events, {'ts': 't', 'is_increase': 'p'})
        events["y"] -= self.minimum_y_value
        events["y"] = events["y"].max() - events["y"]
        target = self.targets[index]
        if self.transform is not None:
            events = self.transform(events)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return events, target

    def __len__(self):
        return len(self.data)

    def _check_exists(self) -> bool:
        return self._is_file_present() and self._folder_contains_at_least_n_files_of_type(
            8000, ".dat"
        )
